using System;
using System.Text;
using org.bouncycastle.math;
using org.bouncycastle.security;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.signers;
using org.bouncycastle.crypto.generators;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.util.encoders;
using org.bouncycastle.util.test;

namespace test.crypto
{

	/**
	* Test based on FIPS 186-2, Appendix 5, an example of DSA.
	*/
	public class DSATest: Test
	{
		internal class MySecureRandom : SecureRandom  
		{
			bool first = true;

			public override void nextBytes(byte[] bytes)
			{
				byte[] k1 = Hex.decode("d5014e4b60ef2ba8b6211b4062ba3224e0427dd3");
				byte[] k2 = Hex.decode("345e8d05c075c3a508df729a1685690e68fcfb8c8117847e89063bca1f85d968fd281540b6e13bd1af989a1fbf17e06462bf511f9d0b140fb48ac1b1baa5bded");

				if (first)
				{
					Array.Copy(k1, 0, bytes, 0, k1.Length);
					first = false;
				}
				else
				{
					Array.Copy(k2, 0, bytes, 0, k2.Length);
				}
			}
		};
		MySecureRandom    random = new MySecureRandom();

		internal class KeySecureRandom : SecureRandom
		{
			public override void nextBytes(byte[] bytes)
			{
				byte[] k = Hex.decode("b5014e4b60ef2ba8b6211b4062ba3224e0427dd3");

				int i;

				for (i = 0; i < (bytes.Length - k.Length); i += k.Length)
				{
					Array.Copy(k, 0, bytes, i, k.Length);
				}

				if (i > bytes.Length)
				{
					Array.Copy(k, 0, bytes, i - k.Length, bytes.Length - (i - k.Length));
				}
				else
				{
					Array.Copy(k, 0, bytes, i, bytes.Length - i);
				}
			}
		};
		KeySecureRandom    keyRandom = new KeySecureRandom();

		BigInteger  pValue = new BigInteger("8df2a494492276aa3d25759bb06869cbeac0d83afb8d0cf7cbb8324f0d7882e5d0762fc5b7210eafc2e9adac32ab7aac49693dfbf83724c2ec0736ee31c80291", 16);
		BigInteger  qValue = new BigInteger("c773218c737ec8ee993b4f2ded30f48edace915f", 16);

		public String getName()
		{
			return "DSA";
		}

		public TestResult perform()
		{
			BigInteger              r = new BigInteger("68076202252361894315274692543577577550894681403");
			BigInteger              s = new BigInteger("1089214853334067536215539335472893651470583479365");
			DSAParametersGenerator  pGen = new DSAParametersGenerator();

			pGen.init(512, 80, random);

			DSAParameters           parameters = pGen.generateParameters();
			DSAValidationParameters pValid = parameters.getValidationParameters();

			if (pValid.getCounter() != 105)
			{
				return new SimpleTestResult(false, getName() + ": Counter wrong");
			}

			if (!pValue.Equals(parameters.getP()) || !qValue.Equals(parameters.getQ()))
			{
				return new SimpleTestResult(false, getName() + ": p or q wrong");
			}

			DSAKeyPairGenerator         dsaKeyGen = new DSAKeyPairGenerator();
			DSAKeyGenerationParameters  genParam = new DSAKeyGenerationParameters(keyRandom, parameters);

			dsaKeyGen.init(genParam);

			AsymmetricCipherKeyPair  pair = dsaKeyGen.generateKeyPair();

			ParametersWithRandom param = new ParametersWithRandom(pair.getPrivate(), keyRandom);

			DSASigner dsa = new DSASigner();

			dsa.init(true, param);

			byte[] message = new BigInteger("968236873715988614170569073515315707566766479517").toByteArray();
			BigInteger[] sig = dsa.generateSignature(message);

			if (!r.Equals(sig[0]))
			{
				return new SimpleTestResult(false, getName()
					+ ": r component wrong." + System.Environment.NewLine
					+ " expecting: " + r + System.Environment.NewLine
					+ " got      : " + sig[0]);
			}

			if (!s.Equals(sig[1]))
			{
				return new SimpleTestResult(false, getName()
					+ ": s component wrong." + System.Environment.NewLine
					+ " expecting: " + s + System.Environment.NewLine
					+ " got      : " + sig[1]);
			}

			dsa.init(false, pair.getPublic());

			if (dsa.verifySignature(message, sig[0], sig[1]))
			{
				return new SimpleTestResult(true, getName() + ": Okay");
			}
			else
			{
				return new SimpleTestResult(false, getName() + ": verification fails");
			}
		}

		[STAThread]
		static void Main(string[] args)
		{
			DSATest         test = new DSATest();
			TestResult      result = test.perform();

			Console.WriteLine(result.ToString());
		}
	}


}
