using System;
using System.Text;
using org.bouncycastle.math;
using org.bouncycastle.math.ec;
using org.bouncycastle.security;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.signers;
using org.bouncycastle.crypto.digests;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.agreement;
using org.bouncycastle.crypto.generators;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.crypto.macs;
using org.bouncycastle.crypto.modes;
using org.bouncycastle.crypto.paddings;
using org.bouncycastle.util.encoders;
using org.bouncycastle.util.test;

namespace test.crypto
{	
	/**
	 * MAC tester - vectors from 
	 * <a href=http://www.itl.nist.gov/fipspubs/fip81.htm>FIP 81</a> and 
	 * <a href=http://www.itl.nist.gov/fipspubs/fip113.htm>FIP 113</a>.
	 */
	public class MacTest: Test
	{
		static byte[]   keyBytes = Hex.decode("0123456789abcdef");
		static byte[]   ivBytes = Hex.decode("1234567890abcdef");

		static byte[]   input = Hex.decode("37363534333231204e6f77206973207468652074696d6520666f7220");

		static byte[]   output1 = Hex.decode("f1d30f68");
		static byte[]   output2 = Hex.decode("58d2e77e");
		static byte[]   output3 = Hex.decode("cd647403");

		public MacTest()
		{
		}

		private bool arraysEqual(
			byte[] a,
			byte[] b)
		{
			if (a.Length != b.Length)
			{
				return false;
			}

			for (int i = 0; i < a.Length; i++)
			{
				if (a[i] != b[i]) return false;
			}

			return true;
		}

		public TestResult perform()
		{
			KeyParameter        key = new KeyParameter(keyBytes);
			BlockCipher         cipher = new DESEngine();
			Mac                 mac = new CBCBlockCipherMac(cipher);

			//
			// standard DAC - zero IV
			//
			mac.init(key);

			mac.update(input, 0, input.Length);

			byte[]  outBytes = new byte[4];

			mac.doFinal(outBytes, 0);

			if (!arraysEqual(outBytes, output1))
			{
				return new SimpleTestResult(false, getName() + ": Failed - expected " + ByteArray2String(Hex.encode(output1)) + " got " + ByteArray2String(Hex.encode(outBytes)));
			}

			//
			// mac with IV.
			//
			ParametersWithIV    param = new ParametersWithIV(key, ivBytes);

			mac.init(param);

			mac.update(input, 0, input.Length);

			outBytes = new byte[4];

			mac.doFinal(outBytes, 0);

			if (!arraysEqual(outBytes, output2))
			{
				return new SimpleTestResult(false, getName() + ": Failed - expected " + ByteArray2String(Hex.encode(output2)) + " got " + ByteArray2String(Hex.encode(outBytes)));
			}
        
			//
			// CFB mac with IV - 8 bit CFB mode
			//
			param = new ParametersWithIV(key, ivBytes);

			mac = new CFBBlockCipherMac(cipher);

			mac.init(param);

			mac.update(input, 0, input.Length);

			outBytes = new byte[4];

			mac.doFinal(outBytes, 0);

			if (!arraysEqual(outBytes, output3))
			{
				return new SimpleTestResult(false, getName() + ": Failed - expected " + ByteArray2String(Hex.encode(output3)) + " got " + ByteArray2String(Hex.encode(outBytes)));
			}

			return new SimpleTestResult(true, getName() + ": Okay");
		}

		public String getName()
		{
			return "Mac";
		}
		private String ByteArray2String(byte[] bs)
		{
			StringBuilder s = new StringBuilder();
			for (int i = 0; i < bs.Length; i++) 
			{
				s.Append(Convert.ToChar(bs[i]));
			}
			return s.ToString();
		}

		[STAThread]
		static void Main(string[] args)
		{
			MacTest    test = new MacTest();
			TestResult result = test.perform();

			Console.WriteLine(result.ToString());
		}
	}

}
