using System;
using System.IO;
using System.Collections;
using org.bouncycastle.io;
using org.bouncycastle.math;
using org.bouncycastle.asn1;
using org.bouncycastle.asn1.x509;

namespace org.bouncycastle.asn1.pkcs
{
	public class PrivateKeyInfo: DEREncodable
	{
		private DERObject               privKey;
		private AlgorithmIdentifier     algId;

		public PrivateKeyInfo(
			AlgorithmIdentifier algId,
			DERObject           privateKey)
		{
			this.privKey = privateKey;
			this.algId = algId;
		}

		public PrivateKeyInfo(
			ASN1Sequence  seq)
		{
			IEnumerator e = seq.getObjects();
			e.MoveNext();
			BigInteger  version = ((DERInteger)e.Current).getValue();
			if (version.intValue() != 0)
			{
				throw new ArgumentException("wrong version for private key info");
			}
			e.MoveNext();
			algId = new AlgorithmIdentifier((ASN1Sequence)e.Current);
			
			try
			{
				e.MoveNext();
				MemoryStream    bIn = new MemoryStream(((ASN1OctetString)e.Current).getOctets());
				DERInputStream          dIn = new DERInputStream(bIn);

				privKey = dIn.readObject();
			}
			catch (IOException ex)
			{
				ex.ToString();
				throw new ArgumentException("Error recoverying private key from sequence");
			}
		}

		public AlgorithmIdentifier getAlgorithmId()
		{
			return algId;
		}

		public DERObject getPrivateKey()
		{
			return privKey;
		}

		/**
		 * write out an RSA private key with it's asscociated information
		 * as described in PKCS8.
		 * <pre>
		 *      PrivateKeyInfo ::= SEQUENCE {
		 *                              version Version,
		 *                              privateKeyAlgorithm AlgorithmIdentifier {{PrivateKeyAlgorithms}},
		 *                              privateKey PrivateKey,
		 *                              attributes [0] IMPLICIT Attributes OPTIONAL 
		 *                          }
		 *      Version ::= INTEGER {v1(0)} (v1,...)
		 *
		 *      PrivateKey ::= OCTET STRING
		 *
		 *      Attributes ::= SET OF Attribute
		 * </pre>
		 */
		public DERObject getDERObject()
		{
			ASN1EncodableVector v = new ASN1EncodableVector();

			v.add(new DERInteger(0));
			v.add(algId);
			v.add(new DEROctetString(privKey));

			return new DERSequence(v);
		}
	}

}