using System;

namespace org.bouncycastle.util.encoders
{
	/**
	 * a buffering class to allow translation from one format to another to
	 * be done in discrete chunks.
	 */
	public class BufferedEncoder
	{
		protected byte[]        buf;
		protected int           bufOff;

		protected Translator    translator;

		/**
		 * @param translator the translator to use.
		 * @param bufSize amount of input to buffer for each chunk.
		 */
		public BufferedEncoder(
			Translator  translator,
			int         bufSize)
		{
			this.translator = translator;

			if ((bufSize % translator.getEncodedBlockSize()) != 0)
			{
				throw new ArgumentException("buffer size not multiple of input block size");
			}

			buf = new byte[bufSize];
			bufOff = 0;
		}

		public int processByte(
			byte        inByte,
			byte[]      outBytes,
			int         outOff)
		{
			int         resultLen = 0;

			buf[bufOff++] = inByte;

			if (bufOff == buf.Length)
			{
				resultLen = translator.encode(buf, 0, buf.Length, outBytes, outOff);
				bufOff = 0;
			}

			return resultLen;
		}

		public int processBytes(
			byte[]      inBytes,
			int         inOff,
			int         len,
			byte[]      outBytes,
			int         outOff)
		{
			if (len < 0)
			{
			throw new ArgumentException("Can't have a negative input length!");
			}

			int resultLen = 0;
			int gapLen = buf.Length - bufOff;

			if (len > gapLen)
			{
				Array.Copy(inBytes, inOff, buf, bufOff, gapLen);

				resultLen += translator.encode(buf, 0, buf.Length, outBytes, outOff);

				bufOff = 0;

				len -= gapLen;
				inOff += gapLen;
				outOff += resultLen;

				int chunkSize = len - (len % buf.Length);

				resultLen += translator.encode(inBytes, inOff, chunkSize, outBytes, outOff);

				len -= chunkSize;
				inOff += chunkSize;
			}

			if (len != 0)
			{
				Array.Copy(inBytes, inOff, buf, bufOff, len);

				bufOff += len;
			}

			return resultLen;
		}
	}

}